/*
 * Trackedit driver for the OSS /dev/dsp output device.
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <unistd.h>
#include <stdlib.h>

#include "drivers.h"

/* Set to 1 to force DSP reset as soon as stopplay() is called ; otherwise */
/* the buffer is emptied before stopping playback. */
#define STOP_DSP_IMMEDIATELY  0

/* ======================================================================== */

int drv_ossdsp_configure(Driver *d, int sample_rate, int n_channels, int sample_size) {
/* 
 * Reconfigures the OSS driver.  This is only called when the driver isn't
 * active, so nothing needs to be done at present beyond what the
 * driver_configure() function does prior to calling this.
 *
 */
  return 1;
}
/* ======================================================================== */

int drv_ossdsp_open(Driver *d) {
/*
 * Opens the specified DSP device and configures it for use with this program.
 * Returns 1 on success.
 *
 */
int fd = open(d->dev.ossdsp.device,O_WRONLY);
audio_buf_info info;
int arg;
int err = 0;

  if (fd<0)
    return(-1);

  /* Request 20 fragments of 2^9=512 bytes (total latency approx 60ms) */
  arg = 0x00140009;
  err = ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &arg)<0;
  arg = d->sample_size;
  err = err ||
        ioctl(fd, SOUND_PCM_WRITE_BITS, &arg)<0 || arg!=d->sample_size;
  arg = d->n_channels;
  err = err ||
        ioctl(fd, SOUND_PCM_WRITE_CHANNELS, &arg)<0 || arg!=d->n_channels;
  arg = d->sample_rate;
  err = err ||
        ioctl(fd, SOUND_PCM_WRITE_RATE, &arg)<0 || arg!=d->sample_rate;

  if (err) {
    close(fd);
    return(0);
  }
  ioctl(fd,SNDCTL_DSP_GETOSPACE,&info);
  d->dev.ossdsp.fd = fd;
  d->dev.ossdsp.bufsize = info.fragsize * info.fragstotal;
  return(1);
}
/* ======================================================================== */

int drv_ossdsp_get_buffer_size(Driver *d) {
/*
 * Returns the optimum buffer size to use with this driver.
 *
 */
  return d->dev.ossdsp.bufsize;
}
/* ======================================================================== */

int drv_ossdsp_close(Driver *d) {
/*
 * Closes the DSP device associated with the given device.  Returns 1 on
 * success.
 *
 */
#if STOP_DSP_IMMEDIATELY
count_info ci;
  /* Find out exactly where the audio is up to */
  ioctl(d->dev.ossdsp.fd,SNDCTL_DSP_GETOPTR,&ci);

  /* Reset the DSP, so any queued sound is dropped */
  ioctl(d->dev.ossdsp.fd,SNDCTL_DSP_RESET,0);

  /* Update the driver `total_written' field to reflect how much audio */
  /* was allowed out of the output (in frames) */
  d->total_written = ci.bytes / ((d->sample_size>>3)*d->n_channels);
#endif

  return close(d->dev.ossdsp.fd)==0;
}
/* ======================================================================== */

int drv_ossdsp_write(Driver *d, char *buf, int size) {
/*
 * Writes `size' bytes from the buffer to the given OSS DSP device.  Returns
 * the number of bytes actually written.
 *
 * The plan: to block until there is <20ms worth of data remaining to be
 * played before committing the new data.  This keeps the write operation at
 * close to realtime.
 *
 */
  return write(d->dev.ossdsp.fd,buf,size);
}
/* ======================================================================== */

int drv_ossdsp_free(Driver *d) {
/*
 * Frees the resources used by the given OSS DSP driver, but not the driver
 * object itself.  Returns 1 on success.
 *
 */
  if (d->dev.ossdsp.device!=NULL)
    free(d->dev.ossdsp.device);
  return 1;
}
/* ======================================================================== */

int drv_ossdsp_init(Driver *d) {
/*
 * Initialises a new OSS DSP driver instance.  Returns 1 on success.
 *
 */
  d->dev.ossdsp.device = NULL;
  d->dev.ossdsp.fd = -1;
  d->dev.ossdsp.bufsize = 0;
  d->open = drv_ossdsp_open;
  d->get_buffer_size = drv_ossdsp_get_buffer_size;
  d->write = drv_ossdsp_write;
  d->close = drv_ossdsp_close;
  d->free = drv_ossdsp_free;

  /* Hardcode for 16 bit, since at this point that's all most OSS cards can */
  /* do.  It can be changed at runtime if the card is known to cope. */
  d->sample_size = 16;
  return 1;
}
/* ======================================================================== */
