/*
 * Module to implement the track editor functions for trackedit.
 *
 * The interface to the driver and other operational aspects (particularly
 * in trackeditor_play() and do_fades()) currently limits the usefulness of
 * this object to input streams consisting of 16 bit signed quantities.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "trackedit.h"

#define TE_BUF_FRAME_COUNT        2048

/* ======================================================================== */
/* Routines to convert to/from dB.  0dB represents the full single-sided
 * scale of the supplied trackeditor.  For a 16 bit signed bitstream, this
 * would be a range from -32768 to +32767.
 */

static float to_dB(Trackeditor *te, float source, int *sign) {
  if (source==0)
    return(-1e6);
  if (sign != NULL)
    *sign = (source > 0)?1:-1;
  return(20.0 * log10(fabs(source)/te->fsd));
}
static float from_dB(Trackeditor *te, float source, int sign) {
  return((sign>0?1:-1)*pow(10,source/20.0)*te->fsd);
}
/* ======================================================================== */

static void do_fades(Trackeditor *te, int bufsize, int currpos, 
  float inslope, float outslope) {
/*
 * Performs fadein/fadeout on the buffer of the given trackeditor.  Currently
 * this routine assumes signed 16 bit samples are in the buffer in little-
 * endian format.
 *
 */
char *c = te->buf;
int d, sample_cx = 0;
int frames_to_in, frames_from_out;
float gain;
  /* If there's no fading to do, just return */
  if ((te->start_frame==-1 || currpos>=te->start_frame+te->fadein_frames) &&
      (te->end_frame==-1 || currpos<te->end_frame-te->fadeout_frames))
    return;

  /* Number of frames to in-point / out-point */
  frames_to_in = te->start_frame+te->fadein_frames - currpos;
  frames_from_out = currpos - (te->end_frame-te->fadeout_frames);
  while (c < te->buf+bufsize) {
    gain = 0;
    if (frames_to_in>0 && te->start_frame != -1)
      gain = inslope*(frames_to_in);
    if (frames_from_out>0 && te->end_frame != -1)
      gain += outslope*(frames_from_out);
    gain = pow(10,gain/20.0);

    for (sample_cx=0;sample_cx<te->channel_count;sample_cx++) {
      d = ((*(c+1) & 0x7f) << 8) + (unsigned char)*c;
      if ((*(c+1) & 0x80)!=0)
        d -= 0x8000;

      /* Calculate new amplitude.  The following essentially does */
      /*   d = from_dB(te,to_dB(te,d,&sign)+gain,sign)            */
      /* but with fewer floating point operations.                */

      d = d*gain;
/*      d = d*(1.0+gain/96.0);*/

      *(c++) = (char)(d & 0xff);
      *(c++) = (char)((d & 0xff00) >> 8);
    }
    frames_to_in--;
    frames_from_out++;
  }
}
/* ======================================================================== */

Trackeditor *trackeditor_new() {
/*
 * Initialises a new Trackeditor instance.
 *
 */
Trackeditor *te = malloc(sizeof(Trackeditor));
  te->input = NULL;
  te->input_filename = NULL;
  te->file_format = AF_FILE_UNKNOWN;
  te->sample_rate = 0;
  te->channel_count = te->sample_size = te->sample_count = 0;
  te->frame_count = 0;
  te->frame_size = 0;
  te->start_frame = te->end_frame = 
    te->fadein_frames = te->fadeout_frames = -1;
  te->fsd = 0;
  te->buf = NULL;
  te->output = NULL;
  return(te);
}
/* ======================================================================== */

void trackeditor_free(Trackeditor *te) {
/*
 * Disposes of the supplied Trackeditor object.  Does *not* dispose or
 * close the driver, since it may be in use elsewhere.
 *
 */
  trackeditor_close_input_file(te);
  if (te->buf!=NULL)
    free(te->buf);
  free(te);
}
/* ======================================================================== */

int trackeditor_get_frame_position(Trackeditor *te) {
/*
 * Returns the current position in frames.
 *
 */
  return afTellFrame(te->input,AF_DEFAULT_TRACK);
}
/* ======================================================================== */

int trackeditor_get_frames_of_time(Trackeditor *te, float sec) {
/*
 * Returns the number of frames required to make up `sec' seconds.
 *
 */
  return (int)(sec*te->sample_rate +0.5);
}
/* ======================================================================== */

float trackeditor_get_time_of_frames(Trackeditor *te, int nframes) {
/*
 * Returns the number of seconds required to play `nframes' frames.
 *
 */
  if (te->sample_rate==0)
    return 0;
  return (float)nframes / te->sample_rate;
}
/* ======================================================================== */

int trackeditor_set_frame_position(Trackeditor *te, int frame) {
/*
 * Sets the current trackeditor position.
 *
 */
AFframecount curr_pos = trackeditor_get_frame_position(te);
  if (frame<0)
    frame = 0;
  if (frame>=te->frame_count)
    frame = te->frame_count-1;
  if (frame != curr_pos)
    afSeekFrame(te->input,AF_DEFAULT_TRACK,frame);
  return 1;
}
/* ======================================================================== */

int trackeditor_set_input_file(Trackeditor *te, char *name) {
/*
 * Attaches the given file to the trackeditor's input.  Returns 1 on success,
 * <= 0 on failure.
 *
 * If `name' is NULL, any open input file is closed.
 *
 */
  trackeditor_close_input_file(te);
  if (name==NULL)
    return(1);
  if ((te->input = afOpenFile(name,"r",AF_NULL_FILESETUP))==AF_NULL_FILEHANDLE) {
    return(0);
  }
  te->frame_count = afGetFrameCount(te->input, AF_DEFAULT_TRACK);
  te->frame_size = afGetFrameSize(te->input, AF_DEFAULT_TRACK, 1);
  te->channel_count = afGetChannels(te->input, AF_DEFAULT_TRACK);
  afGetSampleFormat(te->input, AF_DEFAULT_TRACK, &te->file_format, &te->sample_size);
  te->sample_rate = afGetRate(te->input, AF_DEFAULT_TRACK);

  if (te->output!=NULL) {
    if (driver_configure(te->output,te->sample_rate,te->channel_count,te->sample_size)!=1) {
      fprintf(stderr,"trackeditor_set_input_file(): output sample geometry cannot be set\n");
      trackeditor_close_input_file(te);
      return 0;
    }
  }

  /* This is just here to prevent compiler warnings about from_dB() being */
  /* declared but not used. */
  te->fsd = from_dB(te,0,1);

  te->fsd = pow(2,te->sample_size-1);
  te->input_filename = strdup(name);
  return(1);
}
/* ======================================================================== */

int trackeditor_set_output(Trackeditor *te, Driver *d) {
/*
 * Associates the specified driver with this Trackeditor's output.
 *
 */
  if (te->input!=NULL && d!=NULL) {
    if (driver_configure(d,te->sample_rate,te->channel_count,te->sample_size)!=1) {
      fprintf(stderr,"trackeditor_set_output(): output sample geometry cannot be set\n");
      return 0;
    }
  }
  te->output = d;
  return 1;
}
/* ======================================================================== */

int trackeditor_close_input_file(Trackeditor *te) {
/*
 * Closes any currently active input file.  Returns 1 on success, <=0 on 
 * failure.
 *
 */
  if (te->input != NULL)
    afCloseFile(te->input);
  if (te->input_filename != NULL)
    free(te->input_filename);
  te->input = NULL;
  te->input_filename = NULL;
  te->frame_count = 0;
  te->frame_size = te->channel_count = te->sample_size = 0;
  te->sample_rate = 0;
  te->start_frame = te->end_frame = 
    te->fadein_frames = te->fadeout_frames = -1;
  te->file_format = AF_FILE_UNKNOWN;
  te->fsd = 0;
  return(1);
}
/* ======================================================================== */

int trackeditor_set_play_options(Trackeditor *te, int start_frame, 
  int end_frame, float fadein_time, float fadeout_time) {
/*
 * Sets options for playback.  If start_time or end_time are -1 they
 * are ignored on playback.
 *
 */
int i;
  if (te->input==NULL) {
    fprintf(stderr,"trackeditor_set_play_options(): no input loaded\n");
    return 0;
  }
  if (start_frame<-1) start_frame = -1;
  if (end_frame<-1) end_frame = -1;
  if (start_frame>-1 && end_frame>-1 && start_frame>end_frame) {
    i = start_frame;
    start_frame = end_frame;
    end_frame = i;
  }
  if (start_frame>=te->frame_count) start_frame = te->frame_count-1;
  if (end_frame>=te->frame_count) end_frame = te->frame_count-1;
  te->start_frame = start_frame;
  te->end_frame = end_frame;
  te->fadein_frames = fadein_time<=0?0:trackeditor_get_frames_of_time(te,fadein_time);
  te->fadeout_frames = fadeout_time<=0?0:trackeditor_get_frames_of_time(te,fadeout_time);
  return 1;
}
/* ======================================================================== */

int trackeditor_play(Trackeditor *te, float ofs, float duration) {
/*
 * Plays audio from the input to the output driver.  Playing starts `ofs'
 * seconds from the start of the input track and continues for `duration' 
 * seconds.  If `ofs' is <0, playing starts from the current input position.
 *
 * If the driver is not active, it is opened before writing and closed once
 * playing is complete.  If it's active already, it remains active.
 *
 * Returns 1 on success and 0 on failure.
 *
 */
AFframecount curr_pos = trackeditor_get_frame_position(te);
int total_frames = (int)(te->sample_rate * duration + 0.5);
int bsize, d_active, i;
int frames_in_buffer;
AFframecount ofs_frame;
int ret = 1;
float inslope=0, outslope=0;
  if (te->input==NULL) {
    fprintf(stderr,"trackeditor_play(): no input associated with trackeditor\n");
    return 0;
  }
  if (te->output==NULL) {
    fprintf(stderr,"trackeditor_play(): no driver associated with trackeditor output\n");
    return 0;
  }
  d_active = te->output->active;

  if (ofs<0 && te->start_frame>-1 && curr_pos<te->start_frame) {
    ofs = te->start_frame;
  }

  if (ofs>=0) {
    ofs_frame = (int)(te->sample_rate * ofs + 0.5);
    if (ofs_frame > te->frame_count)
      return(-1);
    /* Only seek if it's absolutely necessary */
    if (ofs_frame != curr_pos)
      afSeekFrame(te->input,AF_DEFAULT_TRACK,ofs_frame);
    curr_pos = ofs_frame;
  }

  /* Set up fadein/fadeout slopes as needed.  Assumes the data is signed */
  /* integer, and there that the smallest "signal" is 0.5 */
  if (te->start_frame+te->fadein_frames>curr_pos) {
    inslope = to_dB(te,0.5,NULL)/te->fadein_frames;
  }
  if (te->end_frame-te->fadeout_frames<=curr_pos) {
    outslope = to_dB(te,0.5,NULL)/te->fadeout_frames;
  }

  bsize = driver_get_buffer_size(te->output);
  if (bsize < 0)
    bsize = TE_BUF_FRAME_COUNT * (te->sample_size >> 3) * te->channel_count;
  frames_in_buffer = bsize / ((te->sample_size>>3)*te->channel_count);
  if (te->buf==NULL) {
    te->buf = malloc(bsize);
  }
  if (!d_active)
    driver_open(te->output);

  i = 0;
  /* Limit playback if required */
  if (te->end_frame>-1) {
    if (curr_pos+total_frames >= te->end_frame)
      total_frames = te->end_frame-curr_pos;
  }
  while (ret && i<total_frames) {
    int nread;
    nread = (i+frames_in_buffer<total_frames)?frames_in_buffer:total_frames-i;
    nread = afReadFrames(te->input,AF_DEFAULT_TRACK,te->buf,nread);
    if (nread != frames_in_buffer)
      bsize = nread * (te->sample_size >> 3) * te->channel_count;
    do_fades(te,bsize,curr_pos+i,inslope,outslope);
    ret = (nread!=0 && driver_write(te->output,te->buf,bsize)==bsize);
    i += nread;
  }
  if (!d_active)
    driver_close(te->output);
  return ret;
}
/* ======================================================================== */

int trackeditor_stopplay(Trackeditor *te) {
/*
 * Stops a current playback.  This forces the output closed immediately. 
 *
 */
  if (te->output==NULL) {
    fprintf(stderr,"trackeditor_stopplay(): no driver associated with trackeditor output\n");
    return 0;
  }
  if (!te->output->active) {
    fprintf(stderr,"trackeditor_stopplay(): output not active\n");
    return 0;
  }
  driver_close(te->output);
  te->start_frame = te->end_frame = -1;
  return 1;
}
/* ======================================================================== */
 